/*
 * $Id: //depot/prod/nvpf/Rinxs/src/java/com/netapp/nvpf/api/model/DatastoreApiService.java#1 $
 *
 * Copyright (c) 2012 NetApp, Inc.
 * All rights reserved.
 */

package com.netapp.nvpf.api.model;

import java.util.List;
import java.util.Map;
import java.util.Set;

import com.netapp.nvpf.api.model.datastore.Datastore;
import com.netapp.nvpf.api.model.datastore.DatastoreInitiateScanValidationResult;
import com.netapp.nvpf.api.model.datastore.DatastoreOptimizationResult;
import com.netapp.nvpf.api.model.datastore.ResizeDatastoreInfo;
import com.netapp.nvpf.api.model.datastore.provisioning.DatastoreCapability;
import com.netapp.nvpf.api.model.datastore.provisioning.ProvisioningDatastoreSpec;
import com.netapp.nvpf.api.model.datastore.provisioning.ProvisioningMessage;
import com.netapp.nvpf.api.model.datastore.provisioning.ProvisioningTargetController;
import com.netapp.nvpf.api.model.datastore.provisioning.ProvisioningTargetControllerDetail;
import com.netapp.nvpf.api.model.storagesystem.DedupeData;
import com.netapp.nvpf.api.model.storagesystem.NasDetails;
import com.netapp.nvpf.api.model.storagesystem.NasDetailsHeader;
import com.netapp.nvpf.api.model.storagesystem.PathOptimizationDialogBean;
import com.netapp.nvpf.api.model.storagesystem.SanDetails;
import com.netapp.nvpf.api.model.storagesystem.SanDetailsHeader;
import com.netapp.nvpf.api.model.virtualmachine.migration.MigrationModel;

/**
 * APIs related to datastores
 * 
 */
public interface DatastoreApiService {

    /**
     * List the basic details of NAS objects (datastores) on a controller
     * 
     * @param controllerId
     *            ID of the controller
     * @return List of NAS detail headers
     * @throws ApiException
     */
    List<NasDetailsHeader> getNasDetailsHeaders(String controllerId, RequestQuery requestQuery) throws ApiException;

    /**
     * List the basic details of SAN objects (VMFS datastores/RDM LUNs) on a
     * controller.
     * 
     * @param controllerId
     *            ID of the controller
     * @return List of SAN detail headers
     * @throws ApiException
     */
    List<SanDetailsHeader> getSanDetailsHeaders(String controllerId, RequestQuery requestQuery) throws ApiException;

    /**
     * Gets full details of a NAS object. The input parameters to this method
     * can be extracted from the NASDetailHeader object. Use this method as a
     * way to get all details for a specific NAS object.
     * 
     * @param datastoreName
     * @param vmfspath
     * @param remotehost
     * @param remotefolder
     * @return NASDetail object
     * @throws ApiException
     */
    NasDetails getNasDetails(String datastoreName, String vmfspath, String remotehost,
        String remotefolder) throws ApiException;

    /**
     * Gets full details of a SAN object. The input parameters to this method
     * can be extracted from the NASDetailHeader object. Use this method as a
     * way to get all details for a specific NAS object.
     * 
     * @param datastoreName
     * @param vmfspath
     * @param remotehost
     * @param remotefolder
     * @return SANDetail object
     * @throws ApiException
     */
    SanDetails getSanDetails(String serial, String durablename) throws ApiException;

    /**
     * Gets full details of a NAS datastore, given the datastore moref. Use this
     * method as a way to get all details for a specific NAS object directly
     * using the moref
     * 
     * @param datastoreMoref
     *            Managed Object reference to datastore as string.
     * 
     * @return NASDetail object
     * @throws ApiException
     */
    NasDetails getNasDetails(String datastoreMoref) throws ApiException;

    /**
     * Returns NAS details of all known controllers.
     *
     * @return Map of controller ID to NAS detail objects.
     *
     * @throws ApiException
     */
    Map<String, Set<NasDetails>> getAllNasDetails() throws ApiException;

    /**
     * Returns SAN details of all known controllers.
     *
     * @return Map of controller ID to SAN detail objects.
     *
     * @throws ApiException
     */
    Map<String, Set<SanDetails>> getAllSanDetails() throws ApiException;


    /**
     * Gets full details of a SAN datastore, given the datastore moref. Use this
     * method as a way to get all details for a specific ANS object directly
     * using the moref.
     * 
     * @param datastoreMoref
     *            Managed Object reference to datastore as string.
     * 
     * @return SANDetail object
     * @throws ApiException
     */
    SanDetails getSanDetails(String datastoreMoref) throws ApiException;

    /**
     * Update deduplication settings for a datastore's underlying flex volume.
     * 
     * @param datastoreMoref
     *            Managed object reference of datastore
     * @param dedupeData
     *            The dedupe settings
     * @throws ApiException
     */
    void updateDedupeOnDataStore(String datastoreMoref, DedupeData dedupeData) throws ApiException;
    
    
    /**
     * Get pathOptimizationDialogBean for direct connected NAS storage system.
     * @param remotePath
     * @param ip
     * @return
     * @throws ApiException
     */
    PathOptimizationDialogBean getPathOptimizationDialogBean(String remotePath, String ip) throws ApiException;

    /**
     * Check Datastore Name Exist or not.
     * @param datastoreName
     * @param session
     * @return Boolean
     * @throws ApiException
     */
    Boolean checkDatastoreNameExists(String datastoreName, String moref) throws ApiException;
    
    /**
     * Fetch the all licensing errors from the request host/datacenter moref field value
     * @param moref
     * @return
     * @throws ApiException
     */
    Map<String, List<ProvisioningMessage>> getLicensingErrors(String moref) throws ApiException;
    
    
    /**
     * Fetch the all list of controllers
     * @param moref
     * @return
     * @throws ApiException
     */
    List<ProvisioningTargetController> getProvisioningTargetController(String datastoreType, String protocol) throws ApiException;
    
    List<ProvisioningTargetController> getProvisioningTargetControllerForPolicy(String datastoreType, String protocol, String policy) throws ApiException;

    /**
     * @param controllerId
     * @param datastoreType
     * @param protocol
     * @return
     */
    ProvisioningTargetControllerDetail getProvisioningTargetControllerDetail(String controllerId, String parentControllerId, 
            String moref, String datastoreType, String protocol) throws ApiException;

    ProvisioningTargetControllerDetail getProvisioningTargetControllerDetailForPolicy(String controllerId, String parentControllerId, String targetMoref, String datastoreType, String protocol, String policy) throws ApiException;

    /**
     * Create datastore.
     * @param controllerId
     * @param pds
     * @return
     * @throws ApiException
     */
    Boolean createDatastore(String controllerId, ProvisioningDatastoreSpec pds) throws ApiException;
    
    /** 
     * is Provisioning
     * @return
     */
    Boolean isProvisioning();

    /**
     * Method for get Datastore Destroy information.
     * @param moref
     * @return
     * @throws ApiException
     */
    Datastore getDatastoreDestroyInfo(String moref) throws ApiException;
    
    /**
     * Get getMountableDataStores associated from particular Host.
     * @param targetMoref
     * @return
     * @throws ApiException 
     */
    List<Datastore> getMountableDatastores(String targetMoref) throws ApiException;

    /**
     * Mount selected datastores on particular host.
     * @param moref
     * @param datastores
     * @throws ApiException
     */
    void mountDatastores(String moref, List<Datastore> datastores) throws ApiException;

    /**
     * Method for destroy datastores.
     * @param moref
     * @param controllerId
     * @throws ApiException
     */
    void destroyDatastore(String moref, String controllerId) throws ApiException;
    
    /**
     * Method for getting Information for space reclaim of datastore. 
     * @param moref
     * @return
     * @throws ApiException
     */
    Datastore getDatastoreReclaimSpaceInfo(String moref) throws ApiException;
    
    /**
     * Method to reclaim space for datastore.
     * @param morefObj
     * @param controllerId
     * @throws ApiException
     */
    void reclaimSpaceForDatastore(String moref, String controllerId) throws ApiException;

    /**
     * @param moref
     * @return
     * @throws ApiException
     */
    ResizeDatastoreInfo getDatastoreResizeInfo(String moref) throws ApiException;

    /**
     * @param moref
     * @param controllerId
     * @param newSize
     * @throws ApiException
     */
    void resizeDatastoreForVMFS(String moref, String controllerId, double newSize) throws ApiException;

    /**
     * @param moref
     * @param controllerId
     * @param newSize
     * @param containerName
     * @throws ApiException
     */
    void resizeDatastoreForNFS(String moref, String controllerId, double newSize, String containerName) throws ApiException;
    
    /**
     * Gets the details of a specified set of datastores
     * @param morefs 
     * @return
     * @throws ApiException
     */
    List<Datastore> getDatastoreDetails(List<String> morefs) throws ApiException;
    
    /**
     * Gets a map with keys as datastore morefs and values as datastores with basic scan details
     * The following fields are populated in the returned datastores: moref, name, isOptimized, isExcludedFromScan,lastScanned.
     * 
     * @param morefs
     * @return
     * @throws ApiException
     */
    Map<String, Datastore> getDatastoreScanDetails(List<String> morefs) throws ApiException;
    
    List<Datastore> getDatastores(MigrationModel migrationModel) throws ApiException;
    
    
    /**
     * Get the list of datastores for scanning. 
     * @param startIndex
     * @param pageSize
     * @param sortedColumn
     * @param ascending
     * @return
     * @throws ApiException
     */
    DatastoreOptimizationResult getDatastoreOptimizationDetails(int startIndex, int pageSize, String sortedColumn, boolean ascending) throws ApiException;
    
    Boolean validateDatastoreNameAgainstVolumeAndLunNames(String datastoreName, String controllerID, String datastoreType, String volumeName, String moref) throws ApiException;
    
    /**
     * Starts an alignment scan for the datastores with the specified moref strings. 
     * Returns a DatastoreInitiateScanValidationResult indicating whether the user has
     * the "Initiate Scan" privilege on all the requested datastores.
     * 
     * @param datastoreMorefs array of datastore moref strings
     * @return a DatastoreInitiateScanValidationResult indicating whether the user has
     * the "Initiate Scan" privilege on all the requested datastores
     * @throws ApiException
     */
    DatastoreInitiateScanValidationResult startAlignmentScan(String[] datastoreMorefs) throws ApiException;
    
    int getNasCount(String controllerId)throws ApiException;
    
    int getSanCount(String controllerId)throws ApiException;
    
    List<String> getPolicies() throws ApiException;

    Boolean isPolicyBasedProvisioningSupported() throws ApiException;
    
    List<DatastoreCapability> getDatastoreCapabilitiesForProfile(final String profileName) throws ApiException;    
}
