/*
 * $Id: //depot/prod/nvpf/Rinxs/src/java/com/netapp/nvpf/api/model/StorageSystemApiService.java#1 $
 *
 * Copyright (c) 2012 NetApp, Inc.
 * All rights reserved.
 */

package com.netapp.nvpf.api.model;

import java.util.List;

import com.netapp.nvpf.api.model.storagesystem.Controller;
import com.netapp.nvpf.api.model.storagesystem.ControllerPrivilegeOpinionBean;
import com.netapp.nvpf.api.model.storagesystem.ControllerTypeEnum;
import com.netapp.nvpf.api.model.storagesystem.DedupeItem;
import com.netapp.nvpf.api.model.storagesystem.LogicalInterface;
import com.netapp.nvpf.api.model.storagesystem.StorageSystemDiscoveryStatus;

/**
 * Storage system related APIs 
 *
 */
public interface StorageSystemApiService {
    /**
     * Gets all clusters
     * @return List of clusters
     * @throws ApiException
     */
    List<Controller> getClusters() throws ApiException;

    /**
     * Gets all Vservers
     * @return List of all Vservers
     * @throws ApiException
     */
    List<Controller> getVServers() throws ApiException;

    /**
     * Gets all vFilers
     * @return List of all vFiler units
     * @throws ApiException
     */
    List<Controller> getVFilers() throws ApiException;

    /**
     * Gets all seven mode storage systems.
     * @return List of seven mode storage systems.
     * @throws ApiException
     */
    List<Controller> getSevenModeStorageSystems() throws ApiException;

    /**
     * Gets all storage systems whose type is unknown
     * @return List of unknown type storage systems 
     * @throws ApiException
     */
    List<Controller> getUnknownStorageSystems() throws ApiException;
    
    /**
     * Gets a list of all storage systems regardless of type
     * @return All storage systems
     * @throws ApiException
     */
    List<Controller> getAllStorageSystems(RequestQuery requestQuery) throws ApiException;

    /**
     * Get the total controller count on the basis of moref and type
     * @param moref
     * @return total count
     * @throws ApiException
     */
    int getControllerCount(String moref, ControllerTypeEnum type) throws ApiException;
    
    /**
     * Get the total controller count on the basis of controllerId and type
     * @param moref
     * @return total count
     * @throws ApiException
     */
    int getRelatedControllerCount(String controllerId, ControllerTypeEnum type) throws ApiException;
    
    /**
     * Get the total interface count on the basis of controllerId
     * @param controllerId
     * @return total count
     * @throws ApiException
     */
    int getRelatedInterfacesCount(String controllerId) throws ApiException;
    
    /**
     * Get all the storage system those are related to the moref field value
     * @param moref
     * @return
     * @throws ApiException
     */
    List<Controller> getAllStorageSystems(String moref, RequestQuery requestQuery) throws ApiException;

    /**
     * Gets the child Vservers of a parent cluster
     * @param controllerId Parent cluster ID
     * @return List of child Vservers
     * @throws ApiException
     */
    List<Controller> getVServers(String controllerId) throws ApiException;

    /**
     * Gets the child vFiler units of a parent seven mode storage system
     * @param controllerId Parent seven mode storage system ID
     * @return List of child vFiler units
     * @throws ApiException
     */
    List<Controller> getVFilers(String controllerId) throws ApiException;

    /**
     * Gets a controller by its ID
     * @param controllerId Controller ID
     * @return Controller object
     * @throws ApiException
     */
    Controller getControllerById(String controllerId) throws ApiException;
    
    /**
     * Gets the list of LIFs for a storage system. Works only for c-mode(clusters and Vservers)
     * @param controllerId ID of the cluster/Vserver
     * @return List of interfaces
     * @throws ApiException
     */
    List<LogicalInterface> getInterfaces(String controllerId, RequestQuery requestQuery) throws ApiException;
    
    /**
     * Gets the list of privileges for a controller user on the specified controller
     * @param controller Object identifying the target controller and the user.
     * @param isNewController Whether this controller is yet to be added to VSC or not.
     * @return An object listing the privileges
     * @throws ApiException
     */
    ControllerPrivilegeOpinionBean getControllerPrivileges(Controller controller,
        Boolean isNewController) throws ApiException;

    /**
     * Gets the provisioning details for a controller.
     * Provisioning details include global provisioning settings and available/usable 
     * resources 9aggregates, volumes interface)
     * 
     * NOTE: The basic details like user name etc will not be returned by this method
     * use the getController() method to get basic details and use this method to get
     * only the provisioning specific details.
     * 
     * @param controllerId ID of the controller
     * @param refershController Whether the details are to re-discovered from ONTAP or not
     * @return Controller object with all provisioning details filled up.
     * @throws ApiException
     */
    Controller getControllerProvisioningDetails(String controllerId, boolean refershController)
        throws ApiException;

    /**
     * Confirms the credentials of a user, passed in the controller object
     * This method extract the user name and password from the incoming controller object
     * and verifies it against the credentials VSC has stored for that controller.
     * 
     * @param controller Controller object with username/password to be validated
     * 
     * @throws ApiException
     */
    void confirmCredentials(Controller controller) throws ApiException;

    /**
     * Gets dedupe and basic storage capacity details for a storage syste,
     * @param controllerId ID of the storage system
     * @return Storage details object
     * @throws ApiException
     */
    DedupeItem getStroageCapacityDetail(String controllerId) throws ApiException;

    /**
     * Discover all ESX hosts and storage system details again.
     * @return A string representing status of the opeation
     * @throws ApiException
     */
    String doDiscovery() throws ApiException;

    /**
     * Discover all ESX hosts and storage system detail those are associated with the given 
     * moref value.
     * @return A string representing status of the opeation
     * @throws ApiException
     */
    String doDiscovery(String moref) throws ApiException;

    /**
     * Adds a new controller to VSC
     * @param controller
     * @return
     * @throws ApiException
     */
    Controller addController(Controller controller) throws ApiException;

    /**
     * Updates a controller
     * @param controller Controller with basic details
     * @param controllerProvisioningSettings Controller with provisioning settings
     * @param controllerProvisioningResources Controller with provisioning resources
     * @throws ApiException
     */
    void updateController(Controller controller, Controller controllerProvisioningSettings,
        Controller controllerProvisioningResources) throws ApiException;

    /**
     * Deletes a controller from VSC repository
     * @param controller
     * @return
     * @throws ApiException
     */
    Boolean deleteController(Controller controller) throws ApiException;

    /**
     * Set Default Credential for controller
     * @param controller
     * @throws ApiException
     */
    void setDefaultControllerCredentials(Controller controller) throws ApiException;

    /**
     * Get Default Credential of controller
     * @return
     * @throws ApiException
     */
    Controller getDefaultControllerCredentials() throws ApiException;
    
    /**
     * Get Storage System Discovery Status
     * @return
     * @throws ApiException
     */
    StorageSystemDiscoveryStatus getStorageSystemDiscoveryStatus() throws ApiException;
}
